// This file may look like C but it is really -*- C++ -*-

// Please see the COPYRIGHT_and_DISCLAIMER file in the top level
// directory of the distribution regarding the provisions under
// which this software is distributed

#include "fmc_defs.h"
#include "global.h"
#include <assert.h>

//
// Function Prototypes
//

SgStatement *   getEndParStmt(SgStatement *s);
void            scanParDo(SgStatement *, SgExpression **);
void            genParBlockCode(SgExpression *, SgStatement *, SgStatement *);
void            doProcessDoStmt(SgStatement *s, SgStatement *, 
				int, SgVarRefExp *);
void            doProsCallStmt(SgStatement *, SgStatement *, 
			       int, SgVarRefExp *);

/*
 * doParallelBlock()
 *
 * Analyze the Parallel Block (PROCESSES/ENDPROCESSES or 
 * PROCESSDO/ENDPROCESSDO).  And generate following calls:
 *
 *  1.  fm_pblock_init() to initialize the process block.
 *  2.  fm_spawn() for each process call.
 *  3.  fm_pblock_reset() to reset the process block.
 *  4.  fm_procname_port_init() to initialize the ports in a process call.
 *  5.  fm_pblock_reset() to reset the process block again.
 *  6.  fm_pblock_context_wait() to wait for the context creation to finish.
 *  7.  fm_port_create() to start port creations.
 *  8.  fm_notify_ports() to do port notifications.
 *  9.  fm_procname_invoke() to invoke processes.
 * 10.  fm_pblock_wait() to wait for the processes to complete.
 *
 */

void doParallelBlock(SgStatement *s, SgStatement **lastStmt)
{
    SgStatement                *endParStmt;
    SgStatement                *line;
    SgStatement                *parLast; 
    SgStatement                *lsp;
    SgExpression               *nslots;
    SgProcessDoStmt            *pd;
    int                        blockNumber;
    int                        gotBlock;
    int                        n_sequential_calls;

    endParStmt               = getEndParStmt(s);
    if (! endParStmt) {
	fm_error(s, "\
No %s statement found in the current procedural block", 
		 (s->variant() == PROCESSES_STAT) ? 
		 "ENDPROCESSES" : "ENDPROCESSDO");
	*lastStmt = s;
	return;
    }

    blockNumber              = 1;
    gotBlock                 = 1;
    nslots                   = c0;
    n_sequential_calls       = 0;

    fmStmtSLL.append(s);
    if (s->variant() == PROCESSES_STAT) {
	fmStmtSLL.append(endParStmt);
	lsp = s;
	for (line = s->lexNext(); lsp != endParStmt; line = lsp->lexNext()) {
	    SgExpression *parDoSlots = c0;
	    //        line->unparsestdout();
//	    printf("\n");
	    switch (line->variant()) {
	      case PROS_STAT:
	      case PROS_STAT_LCTN:
	      case PROS_STAT_SUBM:
		nslots = &(EPlusOne(nslots->copy()));
		lsp = line;
		break;
	      case PROCESS_DO_STAT:
		pd = isSgProcessDoStmt(line);
		scanParDo(line, &parDoSlots);
		parDoSlots = &(EMul(parDoSlots->copy(),
				    loopSize(pd->start(),pd->end(),pd->step())));
//		printf("top level pardoslots = ");
//		parDoSlots->unparsestdout();
//		printf("\n");
		nslots = &(EAdd(nslots->copy(), *parDoSlots));
//		printf("top level nslots = ");
//		nslots->unparsestdout();
//		printf("\n");
		lsp = getEndParStmt(line);
		break;
	      case PROCESSES_END:
		lsp = line;
		break;
	      case PROC_STAT:
		if (n_sequential_calls) {
		    fm_error(line, "\
Only one sequential call is allowed inside a parallel block");
		} else {
		    n_sequential_calls++;
		}
		lsp = line;
		break;
	      default:
		fm_error(line,"Invalid parallel block component: %s",
			 getVariantByName(line->variant()));
		lsp = line;
		break;
	    }
	}
    } else if (s->variant() == PROCESS_DO_STAT) {
	SgExpression *parDoSlots = c0;
	SgProcessDoStmt *pardo = isSgProcessDoStmt(s);

	scanParDo(s, &parDoSlots);
	nslots = &(EMul(*parDoSlots,
			loopSize(pardo->start(),pardo->end(),pardo->step())));
	lsp = endParStmt;
    }
    if (nslots->isInteger()) {
	SgValueExp *n_slots_int = new SgValueExp(nslots->valueInteger());
	genParBlockCode(n_slots_int, s, endParStmt);
    } else {
//	replaceConstantInExpr(*nslots);
	genParBlockCode(nslots, s, endParStmt);
    }
    if (s->variant() == PROCESSES_STAT &&
	currentProcedure->variant() == PROG_HEDR &&
	getEndParStmt(s) == currentProcedure->lastExecutable()) 
    {
	insertShutDownCall(s,1);
    }
    *lastStmt = endParStmt;
}

/*
 * getEndParStmt()
 *
 * Return the ENDPROCESSES or ENDPROCESSDO statement which
 * match the given Parallel statement.
 *
 */

SgStatement *getEndParStmt(SgStatement *s)
{
    SgStatement *lastLineOfCurrentProcedure;

    if (s->variant() == PROCESS_DO_STAT) {
	SgProcessDoStmt *parDo = isSgProcessDoStmt(s);

	if (parDo->endOfLoop()) {
	    parDo->convertLoop();
	}
	return(parDo->lastNodeOfStmt());
    } else {
	// PROCESSES & ENDPROCESSES block
	SgStatement *prog_block = whichProcAmI(s);

	fm_assert(prog_block);
	
	if (prog_block->variant() == PROG_HEDR) {
	    SgProgHedrStmt *prog = isSgProgHedrStmt(prog_block);
        
	    lastLineOfCurrentProcedure = prog->lastNodeOfStmt();
	} else if (prog_block->variant() == PROC_HEDR) {
	    SgProcHedrStmt *proc = isSgProcHedrStmt(prog_block);

	    lastLineOfCurrentProcedure = proc->lastNodeOfStmt();
	} else if (prog_block->variant() == PROS_HEDR) {
	    SgProsHedrStmt *pros = isSgProsHedrStmt(prog_block);
	    
	    lastLineOfCurrentProcedure = pros->lastNodeOfStmt();
	} else if (prog_block->variant() == FUNC_HEDR) {
	    SgFuncHedrStmt *func = isSgFuncHedrStmt(prog_block);
	    
	    lastLineOfCurrentProcedure = func->lastNodeOfStmt();
	} else {
	    fm_error(s, "\
Do not expect a process block to occur in %s",
		     getVariantByName(prog_block->variant()));
	}

	for(SgStatement *line = s->lexNext(); ;line = line->lexNext()) {
	    if (line == lastLineOfCurrentProcedure) {
		fm_error(line, "\
No ENDPROCESSES statement found to match the PROCESSES statement on line");
		// returning the end of procedure as the 
		// last end of the parallel construct.  It is the 
		// simplest reasonable method to recover from this error.
		// The side-effect may be that it may trigger 
		// more compilation errors such as more than one sequential
		// statements in a parallel block...but that should
		// not matter too much.
		return (line);
	    }

	    if (line->variant() == PROCESSES_END) {
		return(line);
	    }
	}
    }
}


/*
 * genParBlockCode()
 * 
 * see the comments on doParallelBlock for more info.
 *
 */

void genParBlockCode (SgExpression *n_slots,
		      SgStatement *start,
		      SgStatement *end)
{
    SgStatement            *line;
    SgSymbol               *fm_pblock_symbol;
    SgSymbol               *fm_buffer_symbol;
    SgSymbol               *fm_ports_symbol; 
    SgVarRefExp            *fm_pblock_exp;
    SgVarRefExp            *fm_buffer_exp;
    SgVarRefExp            *fm_ports_exp;
    SgCallStmt             *callStmt;
    SgStatement            *lsp;
    SgStatement            *serial_call;
    int                    serial_call_found;
    char                   *ident;


    serial_call_found      = 0;
    //
    // Create the Temporary variables used in the process block
    //
    
    ident                  = encodeStr(EC_PARBLOCK);
    fm_pblock_symbol       = new SgVariableSymb(ident, *integer_t);
#if 0
    fm_buffer_symbol       = new SgVariableSymb(encodeStr(EC_BUFFER));
    fm_ports_symbol        = new SgVariableSymb(encodeStr(EC_PORT));
#endif /* 0 */

    fm_pblock_symbol->declareTheSymbol(*currentProcedure);

    if (opt_use_auto_vars) 
      declare_auto_symb(ident, currentProcedure);

    fm_pblock_exp = new SgVarRefExp(*fm_pblock_symbol);

#if 0
    fm_buffer_symbol->setType(*integer_t);
    fm_buffer_symbol->declareTheSymbol(*currentProcedure);
    fm_buffer_exp = new SgVarRefExp(*fm_buffer_symbol);

    fm_ports_symbol->setType(*integer_t);
    fm_ports_symbol->declareTheSymbol(*currentProcedure);
    fm_ports_exp = new SgVarRefExp(*fm_buffer_symbol);
#endif /* 0 */    
    //
    // Generate a call to fm_pblock_init()
    //
    callStmt = new SgCallStmt(*fmPblockInitCallSymb);
    callStmt->addArg(*n_slots);
    callStmt->addArg(*fm_pblock_exp);
    start->insertStmtBefore(*callStmt);
    
    //
    // Phase 0:   spawn 
    //       1:   port initialization
    //       2:   invoking processes
    //
    for (int phase = 0; phase < 3; phase++) {
	lsp = start;
	for (line = start; lsp != end; line = lsp->lexNext()) {
	    switch (line->variant()) {
	      case PROS_STAT:
	      case PROS_STAT_LCTN:
	      case PROS_STAT_SUBM:
#if 0
		doProsCallStmt(line, 
			       start,
			       phase, 
			       fm_pblock_exp, 
			       fm_buffer_exp, 
			       fm_ports_exp);
#endif /* 0 */
		doProsCallStmt(line, 
			       start,
			       phase, 
			       fm_pblock_exp);

		if (phase == 2) {
		    fmStmtSLL.append(line);
		}
		lsp = line;
		break;
	      case PROC_STAT:
		if (! serial_call_found) {
		    serial_call = line;
		    serial_call_found++;
		}
		lsp = line;
		break;
	      case PROCESS_DO_STAT:
#if 0
		doProcessDoStmt(line,
				start,
				phase,
				fm_pblock_exp,
				fm_buffer_exp,
				fm_ports_exp);
#endif /* 0 */
		doProcessDoStmt(line,
				start,
				phase,
				fm_pblock_exp);

		lsp = getEndParStmt(line);
		break;
	      default:
		lsp = line;
		break;
	    }
//	    lsp->unparsestdout();
//	    printf("\n");
	}
	if (phase == 0) {
	    //
	    // Generate a call to fm_pblock_reset()
	    // at the end of phase 0.
	    //
	    callStmt = new SgCallStmt(*fmPblockResetCallSymb);
	    callStmt->addArg(*fm_pblock_exp);
	    start->insertStmtBefore(*callStmt);
	}
	if (phase == 1) {
	    //
	    // Generate the following calls at the end of phase 1:
	    //
	    // 1.  fm_pblock_reset()
	    // 2.  fm_pblock_context_wait()
	    // 3.  fm_port_create()
	    // 4.  fm_notify_ports()
	    //
	    //
	    callStmt = new SgCallStmt(*fmPblockResetCallSymb);
	    callStmt->addArg(*fm_pblock_exp);
	    start->insertStmtBefore(*callStmt);

	    callStmt = new SgCallStmt(*fmPblockContextWaitCallSymb);
	    callStmt->addArg(*fm_pblock_exp);
	    start->insertStmtBefore(*callStmt);
	    
	    callStmt = new SgCallStmt(*fmPortCreateCallSymb);
	    callStmt->addArg(*fm_pblock_exp);
	    start->insertStmtBefore(*callStmt);
	    
	    callStmt = new SgCallStmt(*fmNotifyPortsCallSymb);
	    callStmt->addArg(*fm_pblock_exp);
	    start->insertStmtBefore(*callStmt);
	}
	if (phase == 2) {
	    //
	    // Check whether there is a
	    // serial call inside the block
	    // if so, insert the call here 
	    //
	    //
	    if (serial_call_found) {
		serial_call->deleteStmt();
		start->insertStmtBefore(*serial_call);
	    }
	    //
	    // Generate a call to fm_pblock_wait() at the end of
	    // phase 2.
	    //
	    //
	    callStmt = new SgCallStmt(*fmPblockWaitCallSymb);
	    callStmt->addArg(*fm_pblock_exp);
	    start->insertStmtBefore(*callStmt);
	}
    }
}    

void scanParDo(SgStatement *s, SgExpression **pardo_n_slots)
{
    SgProcessDoStmt *pd;
    SgStatement *endParDo = getEndParStmt(s);
    SgStatement *line;
    SgExpression *sub_n_slots = c0;
    SgExpression *body_n_slots = c0;
    int hasOneProsCall = 0;
    SgStatement *lsp;

    lsp = s;
    for (line = s->lexNext(); lsp != endParDo; line = lsp->lexNext()) {
	switch (line->variant()) {
	  case PROS_STAT_LCTN:
	  case PROS_STAT_SUBM:
	    if (line->variant() == PROS_STAT_LCTN) {
		SgProsCallLctn *lctn = isSgProsCallLctn(line);
		checkVCRef(line, lctn->location());
	    } else {
		SgProsCallSubm *subm = isSgProsCallSubm(line);
		checkVCRef(line, subm->submachine());
	    }
	  case PROS_STAT:
	    if (! hasOneProsCall) {
		hasOneProsCall++;
		body_n_slots = &(EPlusOne(body_n_slots->copy()));
//		printf("body n_slots = ");
//		body_n_slots->unparsestdout();
//		printf("\n");
		*pardo_n_slots = &EAdd((*pardo_n_slots)->copy(), *body_n_slots);
	    } else {
		fm_error(line, "Pardo can only have one call");
	    }
	    lsp = line;
	    break;
	  case PROCESS_DO_STAT:
	    pd = isSgProcessDoStmt(line);
	    scanParDo(line, &sub_n_slots);
//	    body_n_slots = &(EAdd(body_n_slots->copy(), *sub_n_slots));
//	    *pardo_n_slots = &EAdd(pardo_n_slots->copy(), *body_n_slots);
	    *pardo_n_slots = &EMul(*sub_n_slots, 
				   loopSize(pd->start(),pd->end(),pd->step()));
//	    printf("pardo n slots = ");
//	    (*pardo_n_slots)->unparsestdout();
//	    printf("\n");
	    lsp = getEndParStmt(line);
	    
	    break;
	  case CONTROL_END:
	    lsp = line;
	    break;
	  default:
	    fm_error(line, "invalid PROCESSESDO component: %s",
		     getVariantByName(line->variant()));
	    lsp = line;
	    break;
	}
    }
}



