/**********************************************************************
$Id: queue.c,v 1.8 1995/06/14 04:41:09 tuecke Exp $

$Source: /home/compose/Master/FM/src/fm_runtime/queue.c,v $

$Log: queue.c,v $
Revision 1.8  1995/06/14  04:41:09  tuecke
Added support for:
  * bool_t -> fm_bool_t
  * AIX shared libraries
	- makefile.lnk.ibm causes nexus objects to be linked
		directly into libfm.a
	- initwrap.c:fm_init() call nexus_init() and passed the nodes
		onto the real init.  This was done so that the
		nexus_init() would be called from a statically function
  * nexus_create_context() call was changed to use an executable_path
	of "nexus-database:startup-exe", just like the initial
	contexts do.

Revision 1.7  1995/04/17  06:30:28  kohr
Doc. added, moved migration functions to migrate.c.

Revision 1.6  1995/04/06  06:32:32  kohr
Added documentation, cleaned up a bit.


FM runtime support for message queues for FM channels.
**********************************************************************/


#include "fm.h"


static fm_queue_entry_t *alloc_entry();
static void alloc_entries();

/*
_fm_init_queue()
*/

void _fm_init_queue(fm_queue_t *queue)
{
    queue->head.next = &queue->head;
    queue->tail = &queue->head;
    nexus_mutex_init(&queue->lock, (nexus_mutexattr_t *) NULL);
    nexus_cond_init(&queue->cond, (nexus_condattr_t *) NULL);
    queue->n_waiting = 0;
} /* _fm_init_queue() */


/*
_fm_reinit_queue()
*/

void _fm_reinit_queue(fm_queue_t *queue)
{
    queue->head.next = &queue->head;
    queue->tail = &queue->head;
    queue->n_waiting = 0;
} /* _fm_reinit_queue() */


/*
_fm_destroy_queue()
*/

void _fm_destroy_queue(fm_queue_t *queue)
{
} /* _fm_destroy_queue() */

/*
_fm_stash_and_enqueue() : Stash a buffer and enqueue it.
*/

void _fm_stash_and_enqueue(fm_queue_t *queue, nexus_buffer_t *buf)
{
    fm_queue_entry_t *entry;

    entry = alloc_entry();

    nexus_stash_buffer(buf, &entry->buffer);

#ifdef FM_DEBUG
    if (FMDebug(Queue, 2))
	nexus_printf("Enqueue before: queue=%x entry=%x &head=%x "
		     "head.next=%x tail=%x\n",
		     (unsigned long) queue,
		     (unsigned long) entry,
		     (unsigned long) &queue->head,
		     (unsigned long) queue->head.next,
		     (unsigned long) queue->tail);
#endif
    
    nexus_mutex_lock(&queue->lock);
    entry->next = &queue->head;
    queue->tail->next = entry;
    queue->tail = entry;

#ifdef FM_DEBUG
    if (FMDebug(Queue, 1))
	nexus_printf("Enqueue after: queue=%x entry=%x &head=%x "
		     "head.next=%x tail=%x\n",
		     (unsigned long) queue,
		     (unsigned long) entry,
		     (unsigned long) &queue->head,
		     (unsigned long) queue->head.next,
		     (unsigned long) queue->tail);
#if 0
    if (FMDebug(Queue, 2))
	nexus_print_stashed_buffer(&entry->buffer);
#endif
#endif

    if (queue->n_waiting > 0)
	nexus_cond_signal(&queue->cond);
    nexus_mutex_unlock(&queue->lock);
} /* _fm_enqueue() */


/*
_fm_stash_and_enqueue_at_head() : Stash a buffer and enqueue it at the head
of the list (used for forwarded buffers).
*/

void _fm_stash_and_enqueue_at_head(fm_queue_t *queue, nexus_buffer_t *buf)
{
    fm_queue_entry_t *entry;

    entry = alloc_entry();

    nexus_stash_buffer(buf, &entry->buffer);

#ifdef FM_DEBUG
    if (FMDebug(Queue, 2))
	nexus_printf("Enqueue before: queue=%x entry=%x &head=%x "
		     "head.next=%x tail=%x\n",
		     (unsigned long) queue,
		     (unsigned long) entry,
		     (unsigned long) &queue->head,
		     (unsigned long) queue->head.next,
		     (unsigned long) queue->tail);
#endif
    
    nexus_mutex_lock(&queue->lock);
    entry->next = queue->head.next;
    queue->head.next = entry;

    if (queue->tail == &queue->head)
	queue->tail = entry;

#ifdef FM_DEBUG
    if (FMDebug(Queue, 1))
	nexus_printf("Enqueue after: queue=%x entry=%x &head=%x "
		     "head.next=%x tail=%x\n",
		     (unsigned long) queue,
		     (unsigned long) entry,
		     (unsigned long) &queue->head,
		     (unsigned long) queue->head.next,
		     (unsigned long) queue->tail);
#if 0
    if (FMDebug(Queue, 2))
	nexus_print_stashed_buffer(&entry->buffer);
#endif
#endif

    if (queue->n_waiting > 0)
	nexus_cond_signal(&queue->cond);
    nexus_mutex_unlock(&queue->lock);
} /* _fm_stash_and_enqueue_at_head() */


/*
_fm_dequeue()
*/

fm_queue_entry_t *_fm_dequeue(fm_queue_t *queue)
{
    fm_queue_entry_t *entry;
    
    nexus_mutex_lock(&queue->lock);
#ifdef FM_DEBUG
    if (FMDebug(Queue, 1))
	nexus_printf("Dequeue before: queue=%x &head=%x head.next=%x "
		     "tail=%x\n",
		     (unsigned long) queue,
		     (unsigned long) &queue->head,
		     (unsigned long) queue->head.next,
		     (unsigned long) queue->tail);
#endif
    queue->n_waiting++;
    while (queue->head.next == &queue->head)
    {
	nexus_cond_wait(&queue->cond, &queue->lock);
    }
    queue->n_waiting--;

    entry = queue->head.next;
    queue->head.next = entry->next;

    if (entry == queue->tail)
	queue->tail = &queue->head;

#ifdef FM_DEBUG
    if (FMDebug(Queue, 2))
	nexus_printf("Dequeue after: entry=%x &head=%x head.next=%x "
		     "tail=%x\n",
		     (unsigned long) entry,
		     (unsigned long) &queue->head,
		     (unsigned long) queue->head.next,
		     (unsigned long) queue->tail);
#if 0
    if (FMDebug(Queue, 2))
	nexus_print_stashed_buffer(&entry->buffer);
#endif
#endif
    nexus_mutex_unlock(&queue->lock);

    return entry;
} /* _fm_dequeue() */


/*
_fm_queue_empty() : Return true if the queue is empty.
*/

fm_bool_t _fm_queue_empty(fm_queue_t *queue)
{
    fm_bool_t rc;

    nexus_mutex_lock(&queue->lock);
#ifdef FM_DEBUG
    if (FMDebug(Queue, 1))
	nexus_printf("QueueEmtpy: queue=%x &head=%x head.next=%x tail=%x\n",
		     (unsigned long) queue,
		     (unsigned long) &queue->head,
		     (unsigned long) queue->head.next,
		     (unsigned long) queue->tail);
#endif
    rc = queue->head.next == &queue->head;
    nexus_mutex_unlock(&queue->lock);
    return rc;
    
} /* _fm_queue_empty() */


static fm_queue_entry_t *freelist;
static nexus_mutex_t freelist_lock;

/*
_fm_queue_module_init() : Initialize the FM queue stuff...
*/

void _fm_queue_module_init()
{
    nexus_mutex_init(&freelist_lock, (nexus_mutexattr_t *) NULL);
    freelist = (fm_queue_entry_t *) NULL;
    alloc_entries();
} /* _fm_queue_module_init() */


/*
alloc_entries() : Allocate a chunk of queue entries. We assume the freelist
is locked.
*/

static void alloc_entries()
{
    fm_queue_entry_t *entries;
    int i;

    NexusMalloc(alloc_entries(), entries, fm_queue_entry_t *,
		sizeof(fm_queue_entry_t) * FM_QUEUE_ENTRY_ALLOC_SIZE);
    for (i = 0; i < FM_QUEUE_ENTRY_ALLOC_SIZE - 1; i++)
    {
	entries[i].next = &(entries[i+1]);
    }
    entries[i].next = freelist;

    freelist = entries;
} /* alloc_entries() */


/*
alloc_entry() : Allocate a queue entry, out of the free list if possible.
*/

static fm_queue_entry_t *alloc_entry()
{
    fm_queue_entry_t *entry;
    
    nexus_mutex_lock(&freelist_lock);
    if (freelist == (fm_queue_entry_t *) NULL)
	alloc_entries();
    NexusAssert(freelist != (fm_queue_entry_t *) NULL);
    
    entry = freelist;
    freelist = freelist->next;
    nexus_mutex_unlock(&freelist_lock);
    
    entry->next = (fm_queue_entry_t *) NULL;

    return entry;
} /* alloc_entry() */

/*
_fm_free_queue_entry()
*/

void _fm_free_queue_entry(fm_queue_entry_t *entry)
{
    nexus_free_stashed_buffer(&entry->buffer);
    nexus_mutex_lock(&freelist_lock);
    
    entry->next = freelist;
    freelist = entry;

    nexus_mutex_unlock(&freelist_lock);
} /* free_entry() */
