/**********************************************************************
$Id: vc.c,v 1.19 1995/08/12 02:23:43 tuecke Exp $

$Source: /home/compose/Master/FM/src/fm_runtime/vc.c,v $

$Log: vc.c,v $
Revision 1.19  1995/08/12  02:23:43  tuecke
Change nexus-database -> resource_database for newest nexus

Revision 1.18  1995/06/14  04:41:12  tuecke
Added support for:
  * bool_t -> fm_bool_t
  * AIX shared libraries
	- makefile.lnk.ibm causes nexus objects to be linked
		directly into libfm.a
	- initwrap.c:fm_init() call nexus_init() and passed the nodes
		onto the real init.  This was done so that the
		nexus_init() would be called from a statically function
  * nexus_create_context() call was changed to use an executable_path
	of "nexus-database:startup-exe", just like the initial
	contexts do.

Revision 1.17  1995/05/12  20:26:30  tuecke
Updated to new Nexus.
Added AIX shared library support:
	- initwrap.c is necessary on AIX to avoid undefined external symbols
	- Moved sentinel code into here from ../sentinel

Revision 1.16  1995/04/17  06:30:47  kohr
Doc. added, moved migration functions to migrate.c.

Revision 1.15  1995/04/06  06:32:51  kohr
Added documentation, cleaned up a bit.


FM runtime support for virtual computers.
**********************************************************************/


#include "fm.h"
#include "freelist.h"

/* 
The virtual computer table is stored in context-local data under key
vc_key.
*/

typedef struct _vc_table_entry_t
{
    nexus_global_pointer_t gp;
} vc_table_entry_t;

typedef struct _vc_shape_t
{
    int *shape;
    int shape_len;
    int n_nodes;
    int offset;
} vc_shape_t;

typedef int vc_table_offset_t;

typedef struct _vc_table_t
{
    vc_table_entry_t *entries;
    int n_entries;
#if 0
    vc_shape_t *shape;
    int shape_len;
#endif /* 0 */
    nexus_global_pointer_t local_gp;
} vc_table_t;

static nexus_thread_once_t vc_once = NEXUS_THREAD_ONCE_INIT;
static nexus_thread_key_t vc_shape_key;
static nexus_thread_key_t vc_offset_key;

static struct _vc_keys_ {
    nexus_thread_key_t *key;
    nexus_thread_key_destructor_func_t d_func;
} vc_keys[] = {
    { &vc_shape_key, (nexus_thread_key_destructor_func_t)free },
    { &vc_offset_key, (nexus_thread_key_destructor_func_t)free }
};

typedef struct _vc_should_share_response_t
{
    int value;
    fm_synch_t synch;
} vc_should_share_response_t;

static vc_table_t *vc_table;

static void *create_process_thread(void *args);
static void vc_keycreate();
static int compute_vc_offset(int *, int, f_integer_t *, f_integer_t);
static void freelist_init();


/*
_fm_vc_alloc() : Allocate the vc table. This is invoked from NexusBoot().
*/

void _fm_vc_alloc()
{
    NexusMalloc(_fm_vc_alloc(), vc_table, vc_table_t *,
		sizeof(vc_table_t));

    vc_table->entries = (vc_table_entry_t *) NULL;
    vc_table->n_entries = 0;

    freelist_init();

    nexus_global_pointer(&vc_table->local_gp, (void *) vc_table);

} /* _fm_vc_alloc() */


/*
_fm_vc_init() : Initialize the virtual computer. If the list of node_gps is
empty, then this node is a remote node. Just initialize the handlers and
return.

If the list of node_gps is nonempty, then this node is the master.
Distribute the list of gps to the other nodes via the VC_INIT rsr.
*/

void _fm_vc_init(nexus_global_pointer_t *node_gps, int n_node_gps)
{
    nexus_buffer_t mbuf;
    int size, n_elts, obj_n_elts;
    int i;
    nexus_global_pointer_t *gp;
    
    if (n_node_gps == 0)
	return;

    NexusMalloc(_fm_vc_init(), vc_table->entries,
		vc_table_entry_t *, sizeof(vc_table_entry_t) * n_node_gps);

    vc_table->n_entries = n_node_gps;

    vc_table->entries[0].gp = node_gps[0];

    /* 
     * vc_keycreate() is called once for
     * each context.  The master context calls here
     * all other contexts calls vc_keycreate() in
     * _fm_vc_init_handler()  -- Ming --
     */
    nexus_thread_once(&vc_once, vc_keycreate);

    /*
     * Initialize my offset_p
     */
    _fm_vc_table_offset(0);

    _fm_is_node_context = TRUE;

    for (i = 1; i < n_node_gps; i++)
    {
	gp = &(node_gps[i]);

	vc_table->entries[i].gp = *gp;
	
	size = n_elts = 0;

	nexus_init_remote_service_request(&mbuf,
					  gp,
					  FM_VC_INIT_NAME,
					  FM_VC_INIT_HASH);

	/* 
	 * First two integers are:
	 *   -  Boolean value which instantiate _fm_is_node_context
	 *   -  number of global pointers in the VC table
	 *
	 */
	size += nexus_sizeof_int(&mbuf, 1);
	n_elts++;
	
	size += nexus_sizeof_int(&mbuf, 1);
	n_elts++;

	size += nexus_sizeof_global_pointer(&mbuf, node_gps, n_node_gps,
					    &obj_n_elts);
	n_elts += obj_n_elts;

	nexus_set_buffer_size(&mbuf, size, n_elts);

	nexus_put_int(&mbuf, &_fm_is_node_context, 1);
	nexus_put_int(&mbuf, &n_node_gps, 1);
	nexus_put_global_pointer(&mbuf, node_gps, n_node_gps);

#ifdef FM_DEBUG
	if (FMDebug(VC, 3))
	    nexus_printf("_fm_vc_init(): Sending table to %s\n",
			 nexus_global_pointer_string(gp));
#endif
	
	nexus_send_remote_service_request(&mbuf);
    }
} /* _fm_vc_init() */


void _fm_vc_init_handler(void *address, nexus_buffer_t *buffer)
{
    int n_entries;
    int i;

    /* _fm_is_node_context = TRUE; */
    
    nexus_thread_once(&vc_once, vc_keycreate);
    nexus_get_int(buffer, &_fm_is_node_context, 1);
    nexus_get_int(buffer, &n_entries, 1);

#ifdef FM_DEBUG
    if (FMDebug(VC, 3))
	nexus_printf("_fm_vc_init_handler(): n_entries=%d\n", n_entries);
#endif

    NexusMalloc(_fm_vc_init(), vc_table->entries,
		vc_table_entry_t *, sizeof(vc_table_entry_t) * n_entries);

    vc_table->n_entries = n_entries;

    for (i = 0; i < n_entries; i++)
    {
	nexus_get_global_pointer(buffer, &(vc_table->entries[i].gp), 1);
	
#ifdef FM_DEBUG
	if (FMDebug(VC, 4))
	    nexus_printf("_fm_vc_init_handler(): entry %d = %s\n",
			 i,
			 nexus_global_pointer_string(
			 	&(vc_table->entries[i].gp)));
#endif
    }
}


/*
_fm_vc_compute_destination() : Determine the global pointer to the context
for the node corresponding to the given annotation. For now, we only
support local and location annotations. NB, the global pointer return
should *not* be destroyed.
*/

nexus_global_pointer_t _fm_vc_compute_destination(f_integer_t *f_anno,
						  f_integer_t *f_anno_len)
{
    int id, entry_idx;
    nexus_global_pointer_t gp;
    int i, j, prod, offset;
    vc_shape_t *shape_p;
    
    if ((int) (*f_anno_len) == 0 || f_anno[0] == FM_ANNOTATION_LOCAL)
	return vc_table->local_gp;

    NexusAssert(*f_anno_len > 0);

    if (f_anno[0] == FM_ANNOTATION_LOCATION ||
	f_anno[0] == FM_ANNOTATION_SUBMACHINE)
    {
#if 0
        offset = 0;

        for (i = 1; i <= *f_anno_len; i++)
        {
            prod = 1;
            for (j = 1, j <= i - 1; j++)
            {
                prod *= ;
            }
        }
#endif
		/* Can these dead code segments be nuked?--DRK */
#if 0
	if (*f_anno_len > 2)
	{
	    nexus_fatal("_fm_vc_compute_destination(): "
	    	"Relocations of dimension > 1 not yet supported\n");
	}
#endif /* 0 */

	shape_p = (vc_shape_t *) nexus_thread_getspecific(vc_shape_key);
	NexusAssert(shape_p->shape_len == *f_anno_len - 1); 
	NexusAssert(shape_p->n_nodes > 0);

 	id = compute_vc_offset(shape_p->shape, shape_p->shape_len,
			       &f_anno[1], *f_anno_len - 1);
	
	/* Make sure that id falls inside the submachine */
	id = (id - 1) % shape_p->n_nodes + 1;

	NexusAssert(id >= 1);
#ifdef FM_DEBUG
	if (id < 1) 
	    nexus_fatal("_fm_vc_compute_destination(): "
	    	"expect positive location parameter, but get %d\n", id);
#endif /* FM_DEBUG */	

	if (! vc_table->n_entries)
      	    nexus_fatal("vc_table is not initialized\n");

	entry_idx = (id + shape_p->offset - 1) % vc_table->n_entries;

/*	printf("id(%d), offset(%d), entry_idx(%d), size(%d)\n", 
	       id, shape_p->offset, entry_idx, vc_table->n_entries); */

	gp = vc_table->entries[entry_idx].gp;

#ifdef FM_DEBUG
	if (FMDebug(VC, 1))
	    nexus_printf("Location %d maps to entry %d %s\n",
			 id, entry_idx, nexus_global_pointer_string(&gp));
#endif
#if 0
	if (! gp.proto) 
	  nexus_fatal("_fm_vc_compute_destination(): "
	  	"vc_table->entries[%d] has an invalid gp\n", entry_idx);
#endif /* 0 */
	return gp;
    }
#if 0
    else if (f_anno[0] == FM_ANNOTATION_SUBMACHINE)
    {
	nexus_fatal("_fm_vc_compute_destination(): "
		"Submachine annotation not yet supported\n");
    }
#endif /* 0 */
    else
    {
	nexus_fatal("_fm_vc_compute_destination(): "
		    "Unknown annotation type %d\n",
		    f_anno[0]);
    }
} /* _fm_vc_compute_destination() */


/*
_fm_create_process_handler() : Create a new FM process.
The message buffer contains:
     the length of the process name
     the process name
     a global pointer to the pblock creating this process
     the index of the pblock entry for this process
*/

void _fm_create_process_handler(void *address, nexus_buffer_t *buffer)
{
    int process_name_len;
    fm_create_process_info_t *info;
    nexus_thread_t thread;

    info = _fm_alloc_create_process_info();

    nexus_get_int(buffer, &process_name_len, 1);
    nexus_get_char(buffer, info->process_name, process_name_len);
    info->process_name[process_name_len] = '\0';
    nexus_get_global_pointer(buffer, &info->pblock_gp, 1);
    nexus_get_int(buffer, &info->entry_idx, 1);

#ifdef FM_DEBUG
    if (FMDebug(VC, 1))
	nexus_printf("_fm_create_process_handler(): "
		     "process=\"%s\" pblock=%s entry=%d\n",
		     info->process_name,
		     nexus_global_pointer_string(&info->pblock_gp),
		     info->entry_idx);
#endif

	    /* Create a thread to do this, since we'll have to block. */
    nexus_thread_create(&thread, (nexus_thread_attr_t *) NULL,
			create_process_thread, (void *) info);
} /* _fm_create_process_handler() */


static void *create_process_thread(void *args)
{
    fm_create_process_info_t *info;
    vc_should_share_response_t response;
    nexus_global_pointer_t return_gp;
    nexus_global_pointer_t response_gp;
    nexus_global_pointer_t query_context_gp;
    char handler_name[NEXUS_MAX_HANDLER_NAME_LENGTH];
    int handler_hash;
    int size, n_elts, obj_n_elts;
    nexus_buffer_t mbuf;
    int is_node_context;
    int i;

    info = (fm_create_process_info_t *) args;

    _fm_synch_init(&response.synch);
    response.value = -1;
    _fm_synch_increment(&response.synch);

    nexus_global_pointer(&response_gp, (void *) &response);

    /*
     * Here, we would find a context which contains the code for the
     * process. We don't do that yet, instead always use this
     * context.
     */

    query_context_gp = response_gp;

    strcpy(handler_name, FM_SHOULD_SHARE_HANDLER_NAME_PREFIX);
    strcat(handler_name, info->process_name);
    handler_hash = nexus_handler_hash(handler_name);
    nexus_init_remote_service_request(&mbuf,
				      &query_context_gp,
				      handler_name,
				      handler_hash);

    size = nexus_sizeof_global_pointer(&mbuf, &response_gp, 1, &n_elts);
    nexus_set_buffer_size(&mbuf, size, n_elts);
    nexus_put_global_pointer(&mbuf, &response_gp, 1);
    nexus_send_remote_service_request(&mbuf);

#ifdef FM_DEBUG
    if (FMDebug(VC, 1))
	nexus_printf("create_process_thread(): "
		     "waiting for share process return\n");
#endif
    _fm_synch_wait(&response.synch);
#ifdef FM_DEBUG
    if (FMDebug(VC, 1))
	nexus_printf("create_process_thread():  "
		     "share process returns %d _fm_no_threads=%d\n",
		     response.value, _fm_no_threads);
#endif

    /*
     * If the process can share contexts, then we will return the
     * context we queried. Otherwise we create a new context for
     * the process.
     */
    if (response.value && !_fm_no_threads)
    {
	return_gp = query_context_gp;
    }
    else
    {
	int rc;

	nexus_create_context(&query_context_gp,
			     "resource_database:startup-exe",
			     0,
			     &return_gp,
			     &rc,
			     (nexus_create_context_handle_t *) NULL);
	if (rc != 0)
	    nexus_fatal("_fm_create_process_thread(): "
	    		"context creation failed, rc=%d\n",
			rc);
	
	    /* send the VC table to the new context */
	size = n_elts = 0;
	is_node_context = FALSE;

	nexus_init_remote_service_request(&mbuf,
					  &return_gp,
					  FM_VC_INIT_NAME,
					  FM_VC_INIT_HASH);
	
	size += nexus_sizeof_int(&mbuf, 1);
	n_elts++;
	
	size += nexus_sizeof_int(&mbuf, 1);
	n_elts++;
    
	for (i = 0; i < vc_table->n_entries; i++) 
	{
	    size += nexus_sizeof_global_pointer(&mbuf, 
						&vc_table->entries[i].gp,
						1,
						&obj_n_elts);
	    n_elts += obj_n_elts;
	}
	
	nexus_set_buffer_size(&mbuf, size, n_elts);
    
	nexus_put_int(&mbuf, &is_node_context, 1);
	nexus_put_int(&mbuf, &vc_table->n_entries, 1);
	
	for (i = 0; i < vc_table->n_entries; i++) 
	  nexus_put_global_pointer(&mbuf, &vc_table->entries[i].gp, 1);

    
#ifdef FM_DEBUG
	if (FMDebug(VC, 3))
	  nexus_printf("_fm_vc_init(): Sending table to %s\n",
		       nexus_global_pointer_string(&return_gp));
#endif
    
	nexus_send_remote_service_request(&mbuf);
    }

    /*
     * Construct and send the return message. It contains the entry
     * index and the global pointer to the destination context. 
     */
    nexus_init_remote_service_request(&mbuf,
				      &info->pblock_gp,
				      FM_CREATE_PROCESS_RETURN_NAME,
				      FM_CREATE_PROCESS_RETURN_HASH);

    size = n_elts = 0;
    
    size += nexus_sizeof_int(&mbuf, 1);
    n_elts += 1;

    size += nexus_sizeof_global_pointer(&mbuf, &return_gp, 1, &obj_n_elts);
    n_elts += obj_n_elts;

    nexus_set_buffer_size(&mbuf, size, n_elts);

    nexus_put_int(&mbuf, &info->entry_idx, 1);
    nexus_put_global_pointer(&mbuf, &return_gp, 1);
    nexus_send_remote_service_request(&mbuf);

    /*
     * TODO: free up the gps
     */

    _fm_free_create_process_info(info);
    nexus_thread_exit((void *) 0);
} /* _fm_create_process_thread() */


/*
_fm_should_share_return_handler() : Handler for the return from a
ShouldShare request to the context for a new process. The address is that
of a vc_should_share_response_t.

The buffer contains a single integer, the shouldShare value.
*/

void _fm_should_share_return_handler(void *address, nexus_buffer_t *buf)
{
    vc_should_share_response_t *resp;

    resp = (vc_should_share_response_t *) address;

    nexus_get_int(buf, &resp->value, 1);
    _fm_synch_decrement(&resp->synch);
} /* _fm_should_share_return_handler() */


/*
vc_keycreate() : Create keys for the per thread VC shape vector
*/

static void vc_keycreate()
{
    int i;
    
    for (i = 0; i < sizeof(vc_keys) / sizeof(*vc_keys); i++)
        nexus_thread_key_create(vc_keys[i].key, vc_keys[i].d_func);

} /* vc_keycreate */


/*
FM_VC_SHAPE() : Set the shape of the current virtual computer.
*/

void FM_VC_SHAPE(f_integer_t *f_shape, f_integer_t *f_shape_len)
{
    int n_dims = * ((int *) f_shape_len);
    vc_shape_t *shape_p;
    vc_table_offset_t *offset_p;
    int i;
    int n_nodes;
    
    NexusMalloc(FM_VC_SHAPE(), shape_p,
		vc_shape_t *, sizeof(vc_shape_t));

    shape_p->shape_len = n_dims;

    NexusMalloc(FM_VC_SHAPE(), shape_p->shape,
		int *, sizeof(int) * n_dims);
    
    n_nodes = 1;
    for (i = 0; i < n_dims; i++) {
        shape_p->shape[i] = (int) f_shape[i];

	if (shape_p->shape[i] <= 0)
	    nexus_fatal("invalid shape: %d th dimension is %d\n",
			i, shape_p->shape[i]);

	n_nodes *= (int) f_shape[i];
    }

    shape_p->n_nodes = n_nodes;

    offset_p = (vc_table_offset_t *) nexus_thread_getspecific(vc_offset_key);
    if (! offset_p) 
    {
	shape_p->offset = 0;
    } else {
	shape_p->offset = *offset_p;
    }
    NexusAssert(shape_p->offset >= 0);
/*
    printf("offset = %d, n_nodes = %d\n",
	   shape_p->offset,shape_p->n_nodes);
*/
    nexus_thread_setspecific(vc_shape_key, (void *)shape_p);
} /* FM_VC_SHAPE */


/* 
_fm_vc_table_offset() : Set the VC Table offset used for SUBMACHINE
annotations
*/

void _fm_vc_table_offset(int offset)
{

    vc_table_offset_t *offset_p;

    NexusMalloc(_fm_vc_table_offset(),  offset_p,
		vc_table_offset_t *, sizeof(vc_table_offset_t));

#ifdef FM_DEBUG
    if (FMDebug(VC, 3))
        nexus_printf("_fm_vc_table_offset(): setting offset %d\n", offset);
#endif /* FM_DEBUG */
    
    *offset_p = offset;

    nexus_thread_setspecific(vc_offset_key, (void *)offset_p);
}


/* 
FM_CHILD_VC_OFFSET() : Compute the absolute offset for the child into the
VC table given f_anno and f_anno_len.
*/

void FM_CHILD_VC_OFFSET(f_integer_t *f_anno, f_integer_t *f_anno_len, 
		  f_integer_t *vc_offset)
{
    vc_shape_t *shape_p;

    shape_p = (vc_shape_t *) nexus_thread_getspecific(vc_shape_key);
/*    NexusAssert(shape_p); */
    /*
     * If no shape data exsits that means no processors
     * declarations exist so inherit my offset
     *
     */
    if (! shape_p) {
	vc_table_offset_t *offset_p;

	offset_p=(vc_table_offset_t *) nexus_thread_getspecific(vc_offset_key);
	NexusAssert(offset_p);
	*vc_offset = *offset_p;
	return;
    }
    
    if (f_anno[0] != FM_ANNOTATION_SUBMACHINE) {
	*vc_offset = shape_p->offset;
    } else {
	NexusAssert(shape_p->shape_len == *f_anno_len - 1);
	*vc_offset = compute_vc_offset(shape_p->shape, shape_p->shape_len,
				       &f_anno[1], *f_anno_len-1);

	NexusAssert(*vc_offset >= 0);
	*vc_offset = (*vc_offset + shape_p->offset - 1) % shape_p->n_nodes;
    }
}


/*
FM_NUMBER_OF_PROCESSORS() : Return the number of physical processors in the
computation (used for the NumberOfProcessors() intrinsic.
*/

f_integer_t FM_NUMBER_OF_PROCESSORS()
{
    f_integer_t n;
    
    n = vc_table->n_entries;

    return n;
    
} /* FM_NUMBER_OF_PROCESSORS() */


f_integer_t NUMBER_OF_PROCESSORS()
{
    return FM_NUMBER_OF_PROCESSORS();
} /* NUMBER_OF_PROCESSORS() */


static int compute_vc_offset(int *sizes, int n_sizes,
			     f_integer_t *values, f_integer_t n_values)
{
    int sum, prod, i, j;

    sum = 1;

    for (i = 1; i <= n_sizes; i++)
    {
	prod = 1;
	for (j = 1; j <= i - 1; j++)
	{
	    prod *= sizes[j - 1];
	}
	sum += (values[i - 1] - 1) * prod;
    }
    return sum;
}


/*
_fm_vc_shutdown_contexts() : Invoke the shutdown_contexts handler on each
context; incrementing a synch for each. Wait for the synch to go to zero
and return.
*/

void _fm_vc_shutdown_contexts()
{
    fm_synch_t synch;
    int i;
    nexus_buffer_t mbuf;
    nexus_global_pointer_t synch_gp;
    int n_elts, size, obj_n_elts;

    _fm_synch_init(&synch);

    nexus_global_pointer(&synch_gp, (void *) &synch);

    for (i = 1; i < vc_table->n_entries; i++)
    {
	nexus_init_remote_service_request(&mbuf,
					  &vc_table->entries[i].gp,
					  FM_VC_SHUTDOWN_CONTEXTS_NAME,
					  FM_VC_SHUTDOWN_CONTEXTS_HASH);
	size = n_elts = 0;
	size += nexus_sizeof_global_pointer(&mbuf, &synch_gp, 1, &obj_n_elts);
	n_elts += obj_n_elts;

	nexus_set_buffer_size(&mbuf, size, n_elts);

	nexus_put_global_pointer(&mbuf, &synch_gp, 1);

	_fm_synch_increment(&synch);

#ifdef FM_DEBUG
    if (FMDebug(VC, 1))
	nexus_printf("_fm_vc_shutdown_contexts(): Sending shutdown_contexts "
		     "rsr to %d %s\n",
		     i, nexus_global_pointer_string(&vc_table->entries[i].gp));

#endif

	nexus_send_remote_service_request(&mbuf);
    }

    nexus_destroy_global_pointer(&synch_gp);

#ifdef FM_DEBUG
    if (FMDebug(VC, 1))
	nexus_printf("_fm_vc_shutdown_contexts(): Waiting for acks\n");
		     
#endif

    _fm_synch_wait(&synch);
    
#ifdef FM_DEBUG
    if (FMDebug(VC, 1))
	nexus_printf("_fm_vc_shutdown_contexts(): Received all acks\n");
		     
#endif
} /* _fm_vc_shutdown_contexts() */


void _fm_vc_shutdown_contexts_handler(void *address, nexus_buffer_t *buffer)
{
    nexus_global_pointer_t synch_gp;
    nexus_buffer_t mbuf;

    nexus_get_global_pointer(buffer, &synch_gp, 1);

    nexus_init_remote_service_request(&mbuf,
				      &synch_gp,
				      FM_VC_SHUTDOWN_CONTEXTS_ACK_NAME,
				      FM_VC_SHUTDOWN_CONTEXTS_ACK_HASH);

    nexus_send_remote_service_request(&mbuf);

#ifdef FM_DEBUG
    if (FMDebug(VC, 1))
	nexus_printf("_fm_vc_shutdown_contexts_handler(): "
		     "Destroying current context\n");
#endif

    nexus_destroy_current_context(NEXUS_TRUE);
}


void _fm_vc_shutdown_contexts_ack_handler(void *address,
					  nexus_buffer_t *buffer)
{
    fm_synch_t *synch;
    
#ifdef FM_DEBUG
    if (FMDebug(VC, 1))
	nexus_printf("_fm_vc_shutdown_contexts_ack_handler(): Got ack\n");
#endif

    synch = (fm_synch_t *) address;
    SynchMagicCheck(synch);
    _fm_synch_decrement(synch);
}


/*
Use the freelist code for process_info structures.
*/

static void init_create_process_info(fm_create_process_info_t *info)
{
}

static void reinit_create_process_info(fm_create_process_info_t *info)
{
}


FM_FREE_LIST_CODE(create_process_info)


static void freelist_init()
{
    FM_FREE_LIST_INIT(create_process_info);
}
